import click

from shipa.commands import constant
from shipa.utils import validate_map
from shipa.client.types import AppExistsError


@click.group()
def cli_create():
    pass


@cli_create.command("create", short_help=constant.CMD_APP_CREATE)
@click.argument("appname", required=True)
@click.argument("platform", required=False)
@click.option('-d', '--description', help=constant.OPT_APP_CREATE_DESCRIPTION)
@click.option('-f', '--dependency-file', help=constant.OPT_APP_CREATE_DEPENDENCY, multiple=True)
@click.option('-g', '--tag', help=constant.OPT_APP_CREATE_TAG, multiple=True)
@click.option('-p', '--plan', help=constant.OPT_APP_CREATE_PLAN)
@click.option('-t', '--team', help=constant.OPT_APP_CREATE_TEAM, required=True)
@click.option('-F', '--framework', help=constant.OPT_APP_CREATE_FRAMEWORK)
@click.option('-o', '--pool', help=constant.OPT_APP_CREATE_POOL)
@click.option('--ignore-if-exists', help=constant.OPT_APP_CREATE_IGNORE_IF_EXISTS, default=False)
@click.pass_obj
def create(env, appname, platform, description, dependency_file, tag, plan, team, framework, pool, ignore_if_exists):
    """Creates a new app using the given name and platform."""

    with env.client.http as http_client:
        env.client.auth(http_client)
        try:
            if pool is not None:
                framework = pool
                
            env.client.app_create(appname, team=team, framework=framework, platform=platform, description=description,
                                dependency_files=dependency_file, tags=tag, plan=plan)
        except AppExistsError as e:
            print(e)
            if ignore_if_exists:
                pass


@click.group()
def cli_remove():
    pass


@cli_remove.command("remove", short_help=constant.CMD_APP_REMOVE)
@click.argument("appname", required=True)
@click.pass_obj
def remove(env, appname):
    """If the app is bound to any service instance, all binds will be removed before the app gets deleted"""

    with env.client.http as http_client:
        env.client.auth(http_client)
        env.client.app_remove(appname=appname)
        env.client.autoscale_check()
        print('App {0} has been removed!'.format(appname))


@click.group()
def cli_deploy():
    pass


@cli_deploy.command("deploy", short_help=constant.CMD_APP_DEPLOY)
@click.option('-a', '--app', 'appname', help=constant.OPT_APP_DEPLOY_APP, required=True)
@click.option('-d', '--directory', help=constant.OPT_APP_DEPLOY_DIRECTORY, default='.')
@click.option('-i', '--image', help=constant.OPT_APP_DEPLOY_IMAGE)
@click.option('--port', help=constant.OPT_APP_DEPLOY_PORT)
@click.option('--registry-secret', help=constant.OPT_APP_DEPLOY_REGISTRY_SECRET)
@click.option('--registry-user', help=constant.OPT_APP_DEPLOY_REGISTRY_USER)
@click.option('--steps', help=constant.OPT_APP_DEPLOY_STEPS, type=int, default=1, show_default=True)
@click.option('--step-interval', help=constant.OPT_APP_DEPLOY_STEP_INTERVAL)
@click.option('--step-weight', help=constant.OPT_APP_DEPLOY_STEP_WEIGHT)
@click.pass_obj
def deploy(env, appname, image, directory, port, registry_secret, registry_user, steps, step_interval, step_weight):
    """Deploys set of directories to shipa server."""

    with env.client.http as http_client:
        env.client.auth(http_client)
        env.client.app_deploy(appname=appname, image=image, directory=directory, port=port, registry_secret=registry_secret,
                              registry_user=registry_user, steps=steps, step_interval=step_interval,
                              step_weight=step_weight)
        env.client.autoscale_check()
        print('App {0} has been deployed!'.format(appname))


@click.group()
def cli_move():
    pass


@cli_remove.command("move", short_help=constant.CMD_APP_REMOVE)
@click.option('-a', '--app', 'appname', help=constant.OPT_APP_REMOVE_APP, required=True)
@click.option('-p', '--framework', help=constant.OPT_APP_CREATE_FRAMEWORK, required=True)
@click.pass_obj
def move(env, appname, framework):
    """ Moves app to another framework """

    with env.client.http as http_client:
        env.client.auth(http_client)
        env.client.app_move(appname=appname, framework=framework)
        env.client.autoscale_check()
        print('App {0} has been moved!'.format(appname))


cli = click.CommandCollection(sources=[cli_create, cli_remove, cli_deploy, cli_move])



CMD_APP_CREATE = 'create a new app'
OPT_APP_CREATE_DESCRIPTION = 'sets a description for your app'
OPT_APP_CREATE_FRAMEWORK = 'defines which framework your app will be deployed'
OPT_APP_CREATE_POOL = 'defines which pool your app will be deployed to (depracated)'
OPT_APP_CREATE_IGNORE_IF_EXISTS = 'errors will be ignored when creating a resource that already exists'
OPT_APP_CREATE_TEAM = 'describes which team is responsible for the created app'
OPT_APP_CREATE_PLAN = ' parameter defines the plan to be used'
OPT_APP_CREATE_TAG = 'sets a tag to your app. available multiple times'
OPT_APP_CREATE_DEPENDENCY = 'sets inspect file to create app dependencies. available multiple times'
CMD_APP_DEPLOY = 'deploy the source to shipa server'
OPT_APP_DEPLOY_APP = 'specify which app you want to deploy'
OPT_APP_DEPLOY_STEP_INTERVAL = 'single step duration. supported min: m, hour:h, second:s. ex. 1m, 60s, 1h'
OPT_APP_DEPLOY_STEPS = 'steps for canary deployment'
OPT_APP_DEPLOY_DIRECTORY = 'directory to deploy (default .)'
OPT_APP_DEPLOY_IMAGE = 'app image'
OPT_APP_DEPLOY_STEP_WEIGHT = 'deploy step weight'
OPT_APP_DEPLOY_PORT = 'app port (f.e. TCP/8888 or 8888)'
OPT_APP_DEPLOY_REGISTRY_USER = 'username for private registry'
OPT_APP_DEPLOY_REGISTRY_SECRET = 'secret (password) for private registry'
CMD_APP_REMOVE = 'remove a shipa app'
OPT_APP_REMOVE_APP = 'specify which app you want to remove'
CMD_APP_MOVE = 'move a shipa app to another framework'
OPT_APP_MOVE_APP = 'specify which app you want to move'
OPT_APP_MOVE_FRAMEWORK = 'specify which framework you want to move'
CMD_FRAMEWORK_ADD = 'adds a new framework'
OPT_FRAMEWORK_INGRESS = 'ingress controller to use for this framework'
OPT_FRAMEWORK_PUBLIC = 'make framework public (all teams can use it)'
OPT_FRAMEWORK_DEFAULT = 'make framework default (when none is specified during app-create this framework will be used)'
OPT_FRAMEWORK_PROVISIONER = 'provisioner associated to the framework (empty for default docker provisioner)'
OPT_FRAMEWORK_PLAN = 'the plan when adding app to this framework'
OPT_FRAMEWORK_TEAM = 'allowed teams'
OPT_FRAMEWORK_QUOTA_LIMIT = 'quota limit for app when adding it to this framework'
OPT_FRAMEWORK_NAMESPACE = 'kubernetes namespace for this framework'
OPT_FRAMEWORK_DRIVER = 'node drivers accepted in framework. available multiple times'
CMD_FRAMEWORK_REMOVE = 'remove an existing framework.'
CMD_FRAMEWORK_UPDATE = 'updates attributes for a framework.'
CMD_NODE_ADD = 'creates or registers a new node in the cluster'
OPT_NODE_FRAMEWORK = 'framework which will use this node. available multiple times'
OPT_NODE_REGISTER = 'registers an existing docker endpoint, the IaaS won\'t be called.'
OPT_NODE_CA_CERT = 'path to CA file shipa should trust.'
OPT_NODE_CLIENT_CERT = 'path to client TLS certificate file.'
OPT_NODE_CLIENT_KEY = 'path to client TLS key file.'
OPT_NODE_IAAS = 'which iaas provider should be used, if not set shipa will use the default ' \
                'iaas specified in shipa.conf file.'
OPT_NODE_IAASID = 'iaas\'s id.'
OPT_NODE_TEMPLATE = 'A machine template with predefined parameters, additional parameters will override ' \
                    'template ones.'
OPT_NODE_ADDRESS = 'only used if --register option is used. Should point to the endpoint of a working server.'
OPT_NODE_DRIVER = 'which node driver should be used. amazonec2, google, azure, vmwarevsphere, generic, ' \
                  'hyperv, openstack, virtualbox, vmwarefusion'
CMD_NODE_REMOVE = 'removes a node from the cluster'
CMD_NODE_DESTROY = 'destroy node from Iaas'
CMD_NODE_NO_REBALANCE = 'don\'t rebalance containers from removed node'
CMD_CLUSTER_ADD = 'creates a kubernetes cluster definition'
CMD_CLUSTER_ADDRESS = 'address to be used in cluster'
CMD_CLUSTER_TOKEN = 'token used for connection to the cluster'
OPT_CLUSTER_FRAMEWORK = 'framework which will use this cluster'
OPT_CLUSTER_TEAM = 'cluster teams'
OPT_CLUSTER_DEFAULT = 'whether this is the default cluster.'
OPT_CLUSTER_INGRESS_SERVICE = 'specify ingress service type (LoadBalancer/ClusterIP/NodePort)'
OPT_CLUSTER_INGRESS_IP = 'ingress ip address'
OPT_CLUSTER_INGRESS_HTTP_PORT = 'ingress http port'
OPT_CLUSTER_INGRESS_HTTPS_PORT = 'ingress https port'
OPT_CLUSTER_INGRESS_PROTECTED_PORT = 'ingress protected port'
OPT_CLUSTER_INSTALL_CERT_MANAGER = 'install cert manager'
OPT_CLUSTER_CUSTOM = 'KEY=VALUE \n\n' \
                     'custom cluster specific data'
OPT_CLUSTER_PROVISIONER = 'represents a kubernetes cluster'
KUBERNETES_PROVISIONER = 'kubernetes'
OPT_CLUSTER_REMOVE = 'Removes a kubernetes cluster definition'
OPT_ENV_NO_RESTART = 'Sets environment variables without restart the application'
OPT_ENV_APP = 'The name of the app'
OPT_ENV_PRIVATE = 'Private environment variables'
CMD_ENV_UNSET = "Remove environment variables"
CMD_ENV_SET = "Set environment variables"



import click

from shipa.commands import constant


@click.group()
def cli_add():
    pass


@cli_add.command("add", short_help=constant.CMD_FRAMEWORK_ADD)
@click.argument("framework", required=True)
@click.option('-d', '--default', is_flag=True, help=constant.OPT_FRAMEWORK_DEFAULT)
@click.option('-p', '--public', is_flag=True, help=constant.OPT_FRAMEWORK_PUBLIC)
@click.option('--accept-drivers', help=constant.OPT_FRAMEWORK_DRIVER, multiple=True)
@click.option('--app-quota-limit', help=constant.OPT_FRAMEWORK_QUOTA_LIMIT, type=int)
@click.option('--ingress', help=constant.OPT_FRAMEWORK_INGRESS)
@click.option('--kubernetes-namespace', help=constant.OPT_FRAMEWORK_NAMESPACE)
@click.option('--provisioner', help=constant.OPT_FRAMEWORK_PROVISIONER)
@click.option('--plan', help=constant.OPT_FRAMEWORK_PLAN)
@click.option('-t', '--team', help=constant.OPT_FRAMEWORK_TEAM, multiple=True)
@click.pass_obj
def create(env, framework, ingress, default, public, team, accept_drivers, app_quota_limit, provisioner, plan, kubernetes_namespace):
    """Each docker node added by shipa engine belongs to one framework."""

    with env.client.http as http_client:
        env.client.auth(http_client)
        env.client.framework_add(framework=framework, ingress=ingress, default=default, public=public, teams=team, accept_drivers=accept_drivers,
                            app_quota_limit=app_quota_limit, provisioner=provisioner, plan=plan,
                            kubernetes_namespace=kubernetes_namespace)
        print('Framework successfully registered.')


@click.group()
def cli_remove():
    pass


@cli_remove.command("remove", short_help=constant.CMD_FRAMEWORK_REMOVE)
@click.argument("framework", required=True)
@click.pass_obj
def remove(env, framework):
    """Remove an existing framework."""

    with env.client.http as http_client:
        env.client.auth(http_client)
        env.client.framework_remove(framework=framework)
        print('Framework successfully removed.')


@click.group()
def cli_update():
    pass


@cli_update.command("update", short_help=constant.CMD_FRAMEWORK_UPDATE)
@click.argument("framework", required=True)
@click.option('-d', '--default', is_flag=True, help=constant.OPT_FRAMEWORK_DEFAULT)
@click.option('-p', '--public', is_flag=True, help=constant.OPT_FRAMEWORK_PUBLIC)
@click.option('--accept-drivers', help=constant.OPT_FRAMEWORK_DRIVER, multiple=True)
@click.option('--app-quota-limit', help=constant.OPT_FRAMEWORK_QUOTA_LIMIT, type=int)
@click.option('--plan', help=constant.OPT_FRAMEWORK_PLAN)
@click.pass_obj
def update(env, framework, default, public, accept_drivers, app_quota_limit, plan):
    """Updates attributes for a framework."""

    with env.client.http as http_client:
        env.client.auth(http_client)
        env.client.framework_update(framework=framework, default=default, public=public, accept_drivers=accept_drivers,
                               app_quota_limit=app_quota_limit, plan=plan)
        print('Framework successfully updated.')


cli = click.CommandCollection(sources=[cli_add, cli_remove, cli_update])



import unittest

from shipa.client.client import ShipaClient, CONST_TEST_SERVER, CONST_TEST_TOKEN
from shipa.client.http import MockResponse, MockClient


class ShipaEnvTestCase(unittest.TestCase):

    def test_env_set(self):
        try:
            response = MockResponse(code=200)
            http = MockClient(response=response)
            client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)
            client.env_set(app_name="test")

        except Exception as e:
            assert e is None, str(e)

    def test_env_set_failed(self):
        response = MockResponse(code=400, text='failed to set env')
        http = MockClient(response=response)
        client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)

        with self.assertRaises(Exception):
            client.env_set(app_name="test")

    def test_env_unset(self):
        try:
            response = MockResponse(code=200)
            http = MockClient(response=response)
            client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)
            client.env_unset(app_name="test")

        except Exception as e:
            assert e is None, str(e)

    def test_env_unset_failed(self):
        response = MockResponse(code=400, text='failed to remove env')
        http = MockClient(response=response)
        client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)

        with self.assertRaises(Exception):
            client.env_unset(app_name="test")



import os
import tarfile
import tempfile
from collections import defaultdict

import click

from shipa.gitignore import GitIgnore
from shipa.client.types import IngressControllerConfig


class RepositoryFolder(object):
    IGNORE_FILENAME = '.shipaignore'

    def __init__(self, directory, verbose=False):
        assert directory is not None
        assert verbose is not None

        self.directory = directory
        self.verbose = verbose

        ignore_path = os.path.join(directory, self.IGNORE_FILENAME)
        lines = None
        if os.path.isfile(ignore_path) is True:
            with open(ignore_path, 'r') as f:
                lines = f.readlines()
        self.shipa_ignore = GitIgnore(lines or [])

    def create_tarfile(self):

        os.chdir(self.directory)
        if self.verbose:
            print('Create tar archive:')

        def filter(info):
            if info.name.startswith('./.git'):
                return

            filename = info.name[2:]

            if self.shipa_ignore.match(filename):
                if self.verbose:
                    print('IGNORE: ', filename)
                return

            if self.verbose:
                print('OK', filename)
            return info

        f = tempfile.TemporaryFile(suffix='.tar.gz')
        tar = tarfile.open(fileobj=f, mode="w:gz")
        tar.add(name='.',
                recursive=True,
                filter=filter)
        tar.close()
        f.seek(0)
        return f


def parse_step_interval(step_interval):
    if step_interval is None or step_interval == '':
        return 1
    elif step_interval.endswith('s'):
        return int(step_interval[:len(step_interval) - 1])
    elif step_interval.endswith('m'):
        return int(step_interval[:len(step_interval) - 1]) * 60
    elif step_interval.endswith('h'):
        return int(step_interval[:len(step_interval) - 1]) * 60 * 60
    else:
        return step_interval


def parse_port(port):
    port_with_protocol = port.split(":")
    if len(port_with_protocol) == 2:
        return port[0], port[1]

    return "TCP", port


def parse_ingress_config(parameters):
    """parse_ingress_config takes a dict of parameters, groups them by ingress controller name and returns config for
    each controller.

    Args:
        parameters: ...

    Returns:
        A map with IngressControllerConfig for each controller.

    Examples:

        >>> parse_ingress_config({'ingress_ip': ('traefk:10.10.10.10', 'istio:10.10.10.20'), 'ingress_http_port': (
        'traefik:80', 'istio:8080')}) {"traefik": IngressControllerConfig({"ingress_ip":"10.10.10.10",
        "ingress_http_port":"80"),"istio": IngressControllerConfig({"ingress_ip":"10.10.10.20",
        "ingress_http_port":"8080")}
    """

    configs = defaultdict(IngressControllerConfig)
    for name, tuple in parameters.items():
        for value in tuple:
            parsed = value.split(":")
            if len(parsed) != 2:
                raise click.ClickException("invalid %s: %s" % (name, value))
            ingress_name, v = parsed[0], parsed[1]
            configs[ingress_name][name] = v
    return configs


def validate_map(ctx, param, values):
    try:
        map_data = dict()
        for value in values:
            k, v = value.split('=')
            map_data[k] = v
        return map_data
    except ValueError:
        raise click.BadParameter('need to be in format KEY=VALUE')


def validate_env(ctx, param, values):
    try:
        envs = []
        for value in values:
            k, v = value.split('=')
            env = {'name': k, 'value': v}
            envs.append(env)
        return envs
    except ValueError:
        raise click.BadParameter('need to be in format NAME=VALUE')



import unittest

from shipa.client.client import ShipaClient, CONST_TEST_SERVER, CONST_TEST_TOKEN
from shipa.client.http import MockResponse, MockClient


class ShipaFrameworkTestCase(unittest.TestCase):

    def test_framework_create(self):
        try:
            response = MockResponse(code=201)
            http = MockClient(response=response)
            client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)
            client.framework_add(framework="test", ingress="istio")

        except Exception as e:
            assert e is None, str(e)

    def test_framework_create_with_teams(self):
        try:
            response = MockResponse(code=201)
            http = MockClient(response=response)
            client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)
            client.framework_add(framework="test", ingress="istio", teams=("admin", "gaga"))

        except Exception as e:
            assert e is None, str(e)

    def test_framework_create_failed(self):
        response = MockResponse(code=400, text='failed to create a framework')
        http = MockClient(response=response)
        client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)

        with self.assertRaises(Exception):
            client.framework_add(framework="test", ingress="istio")

    def test_framework_remove(self):
        try:
            response = MockResponse(code=200)
            http = MockClient(response=response)
            client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)
            client.framework_remove(framework="test")

        except Exception as e:
            assert e is None, str(e)

    def test_framework_remove_failed(self):
        response = MockResponse(code=400, text='failed to remove a framework')
        http = MockClient(response=response)
        client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)

        with self.assertRaises(Exception):
            client.framework_remove(framework="test")

    def test_framework_update(self):
        try:
            response = MockResponse(code=200)
            http = MockClient(response=response)
            client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)
            client.framework_update(framework="test")

        except Exception as e:
            assert e is None, str(e)

    def test_framework_update_failed(self):
        response = MockResponse(code=400, text='failed to update a framework')
        http = MockClient(response=response)
        client = ShipaClient(server=CONST_TEST_SERVER, client=http, token=CONST_TEST_TOKEN)

        with self.assertRaises(Exception):
            client.framework_update(framework="test")



import click

from shipa.commands import constant
from shipa.utils import validate_env


@click.group()
def cli_set():
    pass


@cli_set.command("set", short_help=constant.CMD_ENV_SET)
@click.argument('env-variables', nargs=-1, required=True, callback=validate_env)
@click.option('-a', '--app', 'app_name', help=constant.OPT_ENV_APP, required=True)
@click.option('-p', '--private', is_flag=True, help=constant.OPT_ENV_PRIVATE)
@click.option('--no-restart', is_flag=True, help=constant.OPT_ENV_NO_RESTART)
@click.pass_obj
def set(env, env_variables, app_name, private, no_restart):
    """Set environment variables"""

    with env.client.http as http_client:
        env.client.auth(http_client)
        env.client.env_set(app_name=app_name, envs=env_variables, private=private, no_restart=no_restart)
        print('Environment variables successfully is successfully set.')


@click.group()
def cli_unset():
    pass


@cli_unset.command("unset", short_help=constant.CMD_ENV_UNSET)
@click.argument('env-variables', nargs=-1, required=True)
@click.option('-a', '--app', 'app_name', help=constant.OPT_ENV_APP, required=True)
@click.option('--no-restart', is_flag=True, help=constant.OPT_ENV_NO_RESTART)
@click.pass_obj
def unset(env, env_variables, app_name, no_restart):
    """Remove environment variables"""

    with env.client.http as http_client:
        env.client.auth(http_client)
        env.client.env_unset(app_name=app_name, envs=env_variables, no_restart=no_restart)
        print('Environment variables successfully removed.')


cli = click.CommandCollection(sources=[cli_set, cli_unset])
